.intel_syntax noprefix
.code32

.equ STACK_SIZE, 65536
.equ MULTIBOOT_MAGIC, 0x1BADB002
.equ MULTIBOOT_FLAGS, 0x00000003

.align 4
multiboot_header:
	.long MULTIBOOT_MAGIC
	.long MULTIBOOT_FLAGS
	.long -(MULTIBOOT_MAGIC + MULTIBOOT_FLAGS)

.global start32
start32:
	# set up known stack
	lea esp, stack + STACK_SIZE
	
	# set up paging structures
	cld
	xor eax, eax	# clear pml4 table
	mov ecx, 1024
	lea edi, pml4Table
	rep stosd
	xor eax, eax	# clear page directory pointer table
	mov ecx, 1024
	lea edi, pageDirPtrTable
	xor eax, eax	# clear page directory table
	mov ecx, 1024
	lea edi, pageDirTable
	rep stosd
	xor eax, eax	# identity map first 1GiB using 2MiB pages
	mov ecx, 512
	lea edi, pageDirTable
nextPDE:
	or eax, 0x00000083		
	mov [edi], eax
	and eax, 0xFFE00000
	add eax, 0x00200000
	add edi, 8
	loop nextPDE
	lea eax, pageDirTable	# set first page directory entry
	or eax, 0x000000003
	mov pageDirPtrTable, eax
	lea eax, pageDirPtrTable	# set first pml4 entry
	or eax, 0x000000003
	mov pml4Table, eax

	# enable PAE	
	mov eax, cr4
	or eax, 0x00000020
	mov cr4, eax
	
	# load pml4 table
	lea eax, pml4Table
	mov cr3, eax

	# set EFER.LME bit
	mov ecx, 0xC0000080
	rdmsr
	or eax, 0x00000100
	wrmsr

	# enable paging
	mov eax, cr0
	or eax, 0x80000000
	mov cr0, eax

	# load gdt
	lgdt [gdtDescriptor]
	
	# set segment selectors
	mov ax, 0x0010
	mov ds, ax
	mov es, ax
	mov fs, ax
	mov gs, ax
	mov ss, ax

	# jump into long mode code
	jmp 0x0008:now64
	
.code64
.extern kmain

now64:
	mov rcx, 0xFFFFFFFF00000000
	and rbx, rcx
	mov rdi, rbx
	call kmain
	hlt

.section .data

GDT:
	.long 0x00000000, 0x00000000	# null
	.long 0x0000FFFF, 0x00EF9A00	# code64
	.long 0x0000FFFF, 0x00EF9200	# data64

gdtDescriptor:
    .word $ - GDT - 1
    .quad GDT
	

.section .bss

.align 4096
.lcomm stack, STACK_SIZE

.align 4096
.lcomm pageDirTable, 4096
.lcomm pageDirPtrTable, 4096
.lcomm pml4Table, 4096

	

